import { assert, CompositeLayer } from '@deck.gl/core';
import { ArcLayer, ColumnLayer, PathLayer, PointCloudLayer, SolidPolygonLayer, PolygonLayer, ScatterplotLayer, TextLayer } from '@deck.gl/layers';
import * as ga from '@geoarrow/geoarrow-js';
import { child } from '@geoarrow/geoarrow-js';
import * as arrow from 'apache-arrow';
import { H3HexagonLayer, TripsLayer } from '@deck.gl/geo-layers';
import { HeatmapLayer } from '@deck.gl/aggregation-layers';
import { Pool, spawn, BlobWorker, Transfer } from 'threads';

// deck.gl-community
// SPDX-License-Identifier: MIT
// Copyright (c) vis.gl contributors
function findGeometryColumnIndex(schema, extensionName, geometryColumnName) {
    const index = schema.fields.findIndex((field) => field.name === geometryColumnName ||
        field.metadata.get("ARROW:extension:name") === extensionName);
    return index !== -1 ? index : null;
}
function isDataInterleavedCoords(data) {
    // TODO: also check 2 or 3d? Float64?
    return arrow.DataType.isFixedSizeList(data.type);
}
function isDataSeparatedCoords(data) {
    // TODO: also check child names? Float64?
    return arrow.DataType.isStruct(data.type);
}
/**
 * Check if the coordinates in a geometry are separate
 * Returns true if the coordinates are separate, false if they are interleaved.
 *
 * The geometry can be point, line, polygon, etc.
 * The function recursively checks for the underlying
 * coordinate data type when it's of type List.
 *
 * If the coordinate type is neither a FixedSizeList nor a Struct,
 * throw an error.
 */
function isGeomSeparate(data) {
    if (arrow.DataType.isList(data.type)) {
        return isGeomSeparate(data.children[0]);
    }
    else if (arrow.DataType.isStruct(data.type)) {
        return true;
    }
    else if (arrow.DataType.isFixedSizeList(data.type)) {
        return false;
    }
    throw new Error(`Unsupported geometry data type: ${data.type}`);
}
/**
 * Convert geoarrow Struct coordinates to FixedSizeList coords
 *
 * The GeoArrow spec allows for either separated or interleaved coords, but at
 * this time deck.gl only supports interleaved.
 */
function convertStructToFixedSizeList(coords) {
    if (isDataInterleavedCoords(coords)) {
        return coords;
    }
    else if (isDataSeparatedCoords(coords)) {
        const nDim = coords.children.length;
        const interleavedCoords = new Float64Array(coords.length * nDim);
        for (let i = 0; i < coords.length; i++) {
            for (let j = 0; j < nDim; j++) {
                interleavedCoords[i * nDim + j] = coords.children[j].values[i];
            }
        }
        const childDataType = new arrow.Float64();
        const dataType = new arrow.FixedSizeList(nDim, new arrow.Field("coords", childDataType));
        const interleavedCoordsData = arrow.makeData({
            type: childDataType,
            length: interleavedCoords.length,
            data: interleavedCoords,
        });
        const data = arrow.makeData({
            type: dataType,
            length: coords.length,
            nullCount: coords.nullCount,
            nullBitmap: coords.nullBitmap,
            child: interleavedCoordsData,
        });
        return data;
    }
    throw new Error(`Unsupported coordinate data type: ${coords.type}`);
}
/**
 * Get LineString Data with interleaved coordinates
 * from the given LineString Data with separated (struct) coordinates.
 */
function getInterleavedLineString(lineStringData) {
    const points = ga.child.getLineStringChild(lineStringData);
    // const coords = getPointChild(points);
    const interleavedPoints = convertStructToFixedSizeList(points);
    return arrow.makeData({
        type: new arrow.List(new arrow.Field("element", interleavedPoints.type)),
        length: lineStringData.length,
        nullCount: lineStringData.nullCount,
        nullBitmap: lineStringData.nullBitmap,
        valueOffsets: lineStringData.valueOffsets,
        offset: lineStringData.offset,
        child: interleavedPoints,
    });
}
/**
 * Get Polygon Data with interleaved coordinates
 * from the given Polygon Data with separated (struct) coordinates.
 */
function getInterleavedPolygon(polygonData) {
    const lineString = ga.child.getPolygonChild(polygonData);
    const interleavedLineString = getInterleavedLineString(lineString);
    return arrow.makeData({
        type: new arrow.List(new arrow.Field("element", interleavedLineString.type)),
        length: polygonData.length,
        nullCount: polygonData.nullCount,
        nullBitmap: polygonData.nullBitmap,
        valueOffsets: polygonData.valueOffsets,
        offset: polygonData.offset,
        child: interleavedLineString,
    });
}
/**
 * A wrapper around a user-provided accessor function
 *
 * For layers like Scatterplot, Path, and Polygon, we automatically handle
 * "exploding" the table when multi-geometry input are provided. This means that
 * the upstream `index` value passed to the user will be the correct row index
 * _only_ for non-exploded data.
 *
 * With this function, we simplify the user usage by automatically converting
 * back from "exploded" index back to the original row index.
 */
function wrapAccessorFunction(objectInfo, userAccessorFunction) {
    const { index, data } = objectInfo;
    let newIndex = index;
    if (data.invertedGeomOffsets !== undefined) {
        newIndex = data.invertedGeomOffsets[index];
    }
    const newObjectData = {
        data: data.data,
        length: data.length,
        attributes: data.attributes,
    };
    const newObjectInfo = {
        index: newIndex,
        data: newObjectData,
        target: objectInfo.target,
    };
    return userAccessorFunction(newObjectInfo);
}
/**
 * Resolve accessor and assign to props object
 *
 * This is useful as a helper function because a scalar prop is set at the top
 * level while a vectorized prop is set inside data.attributes
 *
 */
function assignAccessor(args) {
    const { props, propName, propInput, chunkIdx, geomCoordOffsets } = args;
    if (propInput === undefined) {
        return;
    }
    if (propInput instanceof arrow.Vector) {
        const columnData = propInput.data[chunkIdx];
        if (arrow.DataType.isFixedSizeList(columnData)) {
            assert(columnData.children.length === 1);
            let values = columnData.children[0].values;
            if (geomCoordOffsets) {
                values = expandArrayToCoords(values, columnData.type.listSize, geomCoordOffsets);
            }
            props.data.attributes[propName] = {
                value: values,
                size: columnData.type.listSize,
                // Set to `true` to signify that colors are already 0-255, and deck/luma
                // does not need to rescale
                // https://github.com/visgl/deck.gl/blob/401d624c0529faaa62125714c376b3ba3b8f379f/docs/api-reference/core/attribute-manager.md?plain=1#L66
                normalized: true,
            };
        }
        else if (arrow.DataType.isFloat(columnData)) {
            let values = columnData.values;
            if (geomCoordOffsets) {
                values = expandArrayToCoords(values, 1, geomCoordOffsets);
            }
            props.data.attributes[propName] = {
                value: values,
                size: 1,
            };
        }
    }
    else if (typeof propInput === "function") {
        props[propName] = (object, objectInfo) => {
            // Special case that doesn't have the same parameters
            if (propName === "getPolygonOffset") {
                return propInput(object, objectInfo);
            }
            return wrapAccessorFunction(objectInfo, propInput);
        };
    }
    else {
        props[propName] = propInput;
    }
}
/**
 * Expand an array from "one element per geometry" to "one element per coordinate"
 *
 * @param input: the input array to expand
 * @param size : the number of nested elements in the input array per geometry. So for example, for RGB data this would be 3, for RGBA this would be 4. For radius, this would be 1.
 * @param geomOffsets : an offsets array mapping from the geometry to the coordinate indexes. So in the case of a LineStringArray, this is retrieved directly from the GeoArrow storage. In the case of a PolygonArray, this comes from the resolved indexes that need to be given to the SolidPolygonLayer anyways.
 *
 * @return  {TypedArray} values expanded to be per-coordinate
 */
function expandArrayToCoords(input, size, geomOffsets) {
    const numCoords = geomOffsets[geomOffsets.length - 1];
    // @ts-expect-error
    const outputArray = new input.constructor(numCoords * size);
    // geomIdx is an index into the geomOffsets array
    // geomIdx is also the geometry/table index
    for (let geomIdx = 0; geomIdx < geomOffsets.length - 1; geomIdx++) {
        // geomOffsets maps from the geometry index to the coord index
        // So here we get the range of coords that this geometry covers
        const lastCoordIdx = geomOffsets[geomIdx];
        const nextCoordIdx = geomOffsets[geomIdx + 1];
        // Iterate over this range of coord indices
        for (let coordIdx = lastCoordIdx; coordIdx < nextCoordIdx; coordIdx++) {
            // Iterate over size
            for (let i = 0; i < size; i++) {
                // Copy from the geometry index in `input` to the coord index in
                // `output`
                outputArray[coordIdx * size + i] = input[geomIdx * size + i];
            }
        }
    }
    return outputArray;
}
/**
 * Get a geometry vector with the specified extension type name from the table.
 */
function getGeometryVector(table, geoarrowTypeName) {
    const geometryColumnIdx = findGeometryColumnIndex(table.schema, geoarrowTypeName);
    if (geometryColumnIdx === null) {
        return null;
        // throw new Error(`No column found with extension type ${geoarrowTypeName}`);
    }
    return table.getChildAt(geometryColumnIdx);
}
function getMultiLineStringResolvedOffsets(data) {
    const geomOffsets = data.valueOffsets;
    const lineStringData = ga.child.getMultiLineStringChild(data);
    const ringOffsets = lineStringData.valueOffsets;
    const resolvedRingOffsets = new Int32Array(geomOffsets.length);
    for (let i = 0; i < resolvedRingOffsets.length; ++i) {
        // Perform the lookup into the ringIndices array using the geomOffsets
        // array
        resolvedRingOffsets[i] = ringOffsets[geomOffsets[i]];
    }
    return resolvedRingOffsets;
}
function getPolygonResolvedOffsets(data) {
    const geomOffsets = data.valueOffsets;
    const ringData = ga.child.getPolygonChild(data);
    const ringOffsets = ringData.valueOffsets;
    const resolvedRingOffsets = new Int32Array(geomOffsets.length);
    for (let i = 0; i < resolvedRingOffsets.length; ++i) {
        // Perform the lookup into the ringIndices array using the geomOffsets
        // array
        resolvedRingOffsets[i] = ringOffsets[geomOffsets[i]];
    }
    return resolvedRingOffsets;
}
function getMultiPolygonResolvedOffsets(data) {
    const polygonData = ga.child.getMultiPolygonChild(data);
    const ringData = ga.child.getPolygonChild(polygonData);
    const geomOffsets = data.valueOffsets;
    const polygonOffsets = polygonData.valueOffsets;
    const ringOffsets = ringData.valueOffsets;
    const resolvedRingOffsets = new Int32Array(geomOffsets.length);
    for (let i = 0; i < resolvedRingOffsets.length; ++i) {
        resolvedRingOffsets[i] = ringOffsets[polygonOffsets[geomOffsets[i]]];
    }
    return resolvedRingOffsets;
}
/**
 * Invert offsets so that lookup can go in the opposite direction
 */
function invertOffsets(offsets) {
    const largestOffset = offsets[offsets.length - 1];
    const arrayConstructor = offsets.length < Math.pow(2, 8)
        ? Uint8Array
        : offsets.length < Math.pow(2, 16)
            ? Uint16Array
            : Uint32Array;
    const invertedOffsets = new arrayConstructor(largestOffset);
    for (let arrayIdx = 0; arrayIdx < offsets.length - 1; arrayIdx++) {
        const thisOffset = offsets[arrayIdx];
        const nextOffset = offsets[arrayIdx + 1];
        for (let offset = thisOffset; offset < nextOffset; offset++) {
            invertedOffsets[offset] = arrayIdx;
        }
    }
    return invertedOffsets;
}
// TODO: better typing
function extractAccessorsFromProps(props, excludeKeys) {
    const accessors = {};
    const otherProps = {};
    for (const [key, value] of Object.entries(props)) {
        if (excludeKeys.includes(key)) {
            continue;
        }
        if (key.startsWith("get")) {
            accessors[key] = value;
        }
        else {
            otherProps[key] = value;
        }
    }
    return [accessors, otherProps];
}

// deck.gl-community
// SPDX-License-Identifier: MIT
// Copyright (c) vis.gl contributors
function getPickingInfo({ info, sourceLayer, }, table) {
    // Geometry index as rendered
    let index = info.index;
    // if a Multi- geometry dataset, map from the rendered index back to the
    // feature index
    if (sourceLayer.props.data.invertedGeomOffsets) {
        index = sourceLayer.props.data.invertedGeomOffsets[index];
    }
    const recordBatchIdx = sourceLayer.props.recordBatchIdx;
    const tableOffsets = sourceLayer.props.tableOffsets;
    const batch = table.batches[recordBatchIdx];
    const row = batch.get(index);
    if (row === null) {
        return info;
    }
    const currentBatchOffset = tableOffsets[recordBatchIdx];
    // Update index to be _global_ index, not within the specific record batch
    index += currentBatchOffset;
    return {
        ...info,
        index,
        object: row,
    };
}
// This is vendored from Arrow JS because it's a private API
function computeChunkOffsets(chunks) {
    return chunks.reduce((offsets, chunk, index) => {
        offsets[index + 1] = offsets[index] + chunk.length;
        return offsets;
    }, new Uint32Array(chunks.length + 1));
}

// deck.gl-community
// SPDX-License-Identifier: MIT
// Copyright (c) vis.gl contributors
function validateAccessors(props, table) {
    const vectorAccessors = [];
    const colorVectorAccessors = [];
    for (const [accessorName, accessorValue] of Object.entries(props)) {
        // Is it an accessor
        if (accessorName.startsWith("get")) {
            // Is it a vector accessor
            if (accessorValue instanceof arrow.Vector) {
                vectorAccessors.push(accessorValue);
                // Is it a color vector accessor
                if (accessorName.endsWith("Color")) {
                    colorVectorAccessors.push(accessorValue);
                }
            }
        }
    }
    validateVectorAccessors(table, vectorAccessors);
    for (const colorVectorAccessor of colorVectorAccessors) {
        validateColorVector(colorVectorAccessor);
    }
}
/**
 * Provide validation for accessors provided
 *
 * - Assert that all vectors have the same number of chunks as the main table
 * - Assert that all chunks in each vector have the same number of rows as the
 *   relevant batch in the main table.
 *
 */
function validateVectorAccessors(table, vectorAccessors) {
    // Check the same number of chunks as the table's batches
    for (const vectorAccessor of vectorAccessors) {
        assert(table.batches.length === vectorAccessor.data.length);
    }
    // Check that each table batch/vector data has the same number of rows
    for (const vectorAccessor of vectorAccessors) {
        for (let i = 0; i < table.batches.length; i++) {
            assert(table.batches[i].numRows === vectorAccessor.data[i].length);
        }
    }
}
function validateColorVector(vector) {
    // Assert the color vector is a FixedSizeList
    assert(arrow.DataType.isFixedSizeList(vector.type));
    // Assert it has 3 or 4 values
    assert(vector.type.listSize === 3 || vector.type.listSize === 4);
    // Assert the child type is an integer
    assert(arrow.DataType.isInt(vector.type.children[0]));
    // Assert the child type is a Uint8
    // @ts-ignore
    // Property 'type' does not exist on type 'Int_<Ints>'. Did you mean 'TType'?
    assert(vector.type.children[0].type.bitWidth === 8);
}

// deck.gl-community
// SPDX-License-Identifier: MIT
// Copyright (c) vis.gl contributors
// Remove data from the upstream default props
const { data: _data$a, getSourcePosition: _getSourcePosition, getTargetPosition: _getTargetPosition, ..._defaultProps$8 } = ArcLayer.defaultProps;
// Default props added by us
const ourDefaultProps$a = {
    _validate: true,
};
// @ts-expect-error Type error in merging default props with ours
const defaultProps$a = {
    ..._defaultProps$8,
    ...ourDefaultProps$a,
};
class GeoArrowArcLayer extends CompositeLayer {
    static defaultProps = defaultProps$a;
    static layerName = "GeoArrowArcLayer";
    getPickingInfo(params) {
        return getPickingInfo(params, this.props.data);
    }
    renderLayers() {
        return this._renderLayersPoint();
    }
    _renderLayersPoint() {
        const { data: table, getSourcePosition: sourcePosition, getTargetPosition: targetPosition, } = this.props;
        if (this.props._validate) {
            validateAccessors(this.props, table);
            // Note: below we iterate over table batches anyways, so this layer won't
            // work as-is if data/table is null
            assert(ga.vector.isPointVector(sourcePosition));
            assert(ga.vector.isPointVector(targetPosition));
        }
        // Exclude manually-set accessors
        const [accessors, otherProps] = extractAccessorsFromProps(this.props, [
            "getSourcePosition",
            "getTargetPosition",
        ]);
        const tableOffsets = computeChunkOffsets(table.data);
        const layers = [];
        for (let recordBatchIdx = 0; recordBatchIdx < table.batches.length; recordBatchIdx++) {
            let sourceData = sourcePosition.data[recordBatchIdx];
            if (isGeomSeparate(sourceData)) {
                sourceData = convertStructToFixedSizeList(sourceData);
            }
            const sourceValues = child.getPointChild(sourceData).values;
            let targetData = targetPosition.data[recordBatchIdx];
            if (isGeomSeparate(targetData)) {
                targetData = convertStructToFixedSizeList(targetData);
            }
            const targetValues = child.getPointChild(targetData).values;
            const props = {
                // Note: because this is a composite layer and not doing the rendering
                // itself, we still have to pass in our defaultProps
                ...ourDefaultProps$a,
                ...otherProps,
                // used for picking purposes
                recordBatchIdx,
                tableOffsets,
                id: `${this.props.id}-geoarrow-arc-${recordBatchIdx}`,
                data: {
                    // @ts-expect-error passed through to enable use by function accessors
                    data: table.batches[recordBatchIdx],
                    length: sourceData.length,
                    attributes: {
                        getSourcePosition: {
                            value: sourceValues,
                            size: sourceData.type.listSize,
                        },
                        getTargetPosition: {
                            value: targetValues,
                            size: targetData.type.listSize,
                        },
                    },
                },
            };
            for (const [propName, propInput] of Object.entries(accessors)) {
                assignAccessor({
                    props,
                    propName,
                    propInput,
                    chunkIdx: recordBatchIdx,
                });
            }
            const layer = new ArcLayer(this.getSubLayerProps(props));
            layers.push(layer);
        }
        return layers;
    }
}

// deck.gl-community
// SPDX-License-Identifier: MIT
// Copyright (c) vis.gl contributors
/**
 * Enum holding GeoArrow extension type names
 */
var EXTENSION_NAME;
(function (EXTENSION_NAME) {
    EXTENSION_NAME["POINT"] = "geoarrow.point";
    EXTENSION_NAME["LINESTRING"] = "geoarrow.linestring";
    EXTENSION_NAME["POLYGON"] = "geoarrow.polygon";
    EXTENSION_NAME["MULTIPOINT"] = "geoarrow.multipoint";
    EXTENSION_NAME["MULTILINESTRING"] = "geoarrow.multilinestring";
    EXTENSION_NAME["MULTIPOLYGON"] = "geoarrow.multipolygon";
})(EXTENSION_NAME || (EXTENSION_NAME = {}));

// deck.gl-community
// SPDX-License-Identifier: MIT
// Copyright (c) vis.gl contributors
// Remove data and getPosition from the upstream default props
const { data: _data$9, getPosition: _getPosition$4, ..._defaultProps$7 } = ColumnLayer.defaultProps;
// Default props added by us
const ourDefaultProps$9 = {
    _validate: true,
};
// @ts-expect-error Type error in merging default props with ours
const defaultProps$9 = {
    ..._defaultProps$7,
    ...ourDefaultProps$9,
};
/**
 * Render extruded cylinders (tessellated regular polygons) at given
 * coordinates.
 */
class GeoArrowColumnLayer extends CompositeLayer {
    static defaultProps = defaultProps$9;
    static layerName = "GeoArrowColumnLayer";
    getPickingInfo(params) {
        return getPickingInfo(params, this.props.data);
    }
    renderLayers() {
        const { data: table } = this.props;
        const pointVector = getGeometryVector(table, EXTENSION_NAME.POINT);
        if (pointVector !== null) {
            return this._renderLayersPoint(pointVector);
        }
        const geometryColumn = this.props.getPosition;
        if (geometryColumn !== undefined &&
            ga.vector.isPointVector(geometryColumn)) {
            return this._renderLayersPoint(geometryColumn);
        }
        throw new Error("getPosition not GeoArrow point");
    }
    _renderLayersPoint(geometryColumn) {
        const { data: table } = this.props;
        if (this.props._validate) {
            assert(ga.vector.isPointVector(geometryColumn));
            validateAccessors(this.props, table);
        }
        // Exclude manually-set accessors
        const [accessors, otherProps] = extractAccessorsFromProps(this.props, [
            "getPosition",
        ]);
        const tableOffsets = computeChunkOffsets(table.data);
        const layers = [];
        for (let recordBatchIdx = 0; recordBatchIdx < table.batches.length; recordBatchIdx++) {
            let geometryData = geometryColumn.data[recordBatchIdx];
            if (isGeomSeparate(geometryData)) {
                geometryData = convertStructToFixedSizeList(geometryData);
            }
            const flatCoordsData = ga.child.getPointChild(geometryData);
            const flatCoordinateArray = flatCoordsData.values;
            const props = {
                // Note: because this is a composite layer and not doing the rendering
                // itself, we still have to pass in our defaultProps
                ...ourDefaultProps$9,
                ...otherProps,
                // used for picking purposes
                recordBatchIdx,
                tableOffsets,
                id: `${this.props.id}-geoarrow-column-${recordBatchIdx}`,
                data: {
                    // @ts-expect-error passed through to enable use by function accessors
                    data: table.batches[recordBatchIdx],
                    length: geometryData.length,
                    attributes: {
                        getPosition: {
                            value: flatCoordinateArray,
                            size: geometryData.type.listSize,
                        },
                    },
                },
            };
            for (const [propName, propInput] of Object.entries(accessors)) {
                assignAccessor({
                    props,
                    propName,
                    propInput,
                    chunkIdx: recordBatchIdx,
                });
            }
            const layer = new ColumnLayer(this.getSubLayerProps(props));
            layers.push(layer);
        }
        return layers;
    }
}

// deck.gl-community
// SPDX-License-Identifier: MIT
// Copyright (c) vis.gl contributors
// Remove data from the upstream default props
const { data: _data$8, getHexagon: _getHexagon, ..._defaultProps$6 } = H3HexagonLayer.defaultProps;
// Default props added by us
const ourDefaultProps$8 = {
    _validate: true,
};
const defaultProps$8 = {
    ..._defaultProps$6,
    ...ourDefaultProps$8,
};
class GeoArrowH3HexagonLayer extends CompositeLayer {
    static defaultProps = defaultProps$8;
    static layerName = "GeoArrowH3HexagonLayer";
    getPickingInfo(params) {
        return getPickingInfo(params, this.props.data);
    }
    renderLayers() {
        return this._renderLayersPoint();
    }
    _renderLayersPoint() {
        const { data: table, getHexagon: hexagonColumn } = this.props;
        if (this.props._validate) {
            validateAccessors(this.props, table);
        }
        // Exclude manually-set accessors
        const [accessors, otherProps] = extractAccessorsFromProps(this.props, [
            "getHexagon",
        ]);
        const tableOffsets = computeChunkOffsets(table.data);
        const layers = [];
        for (let recordBatchIdx = 0; recordBatchIdx < table.batches.length; recordBatchIdx++) {
            const hexData = hexagonColumn.data[recordBatchIdx];
            const hexValues = hexData.values;
            const props = {
                // Note: because this is a composite layer and not doing the rendering
                // itself, we still have to pass in our defaultProps
                ...ourDefaultProps$8,
                ...otherProps,
                // used for picking purposes
                recordBatchIdx,
                tableOffsets,
                id: `${this.props.id}-geoarrow-arc-${recordBatchIdx}`,
                data: {
                    // @ts-expect-error passed through to enable use by function accessors
                    data: table.batches[recordBatchIdx],
                    length: hexData.length,
                    attributes: {
                        getHexagon: {
                            value: hexValues,
                            // h3 cells should always be 15 characters...?
                            size: 15,
                        },
                    },
                },
            };
            for (const [propName, propInput] of Object.entries(accessors)) {
                assignAccessor({
                    props,
                    propName,
                    propInput,
                    chunkIdx: recordBatchIdx,
                });
            }
            const layer = new H3HexagonLayer(this.getSubLayerProps(props));
            layers.push(layer);
        }
        return layers;
    }
}

// deck.gl-community
// SPDX-License-Identifier: MIT
// Copyright (c) vis.gl contributors
// Remove data from the upstream default props
const { data: _data$7, getPosition: _getPosition$3, ..._defaultProps$5 } = HeatmapLayer.defaultProps;
// Default props added by us
const ourDefaultProps$7 = {
    _validate: true,
};
// @ts-expect-error Type error in merging default props with ours
const defaultProps$7 = {
    ..._defaultProps$5,
    ...ourDefaultProps$7,
};
class GeoArrowHeatmapLayer extends CompositeLayer {
    static defaultProps = defaultProps$7;
    static layerName = "GeoArrowHeatmapLayer";
    renderLayers() {
        const { data: table } = this.props;
        if (this.props.getPosition !== undefined) {
            const geometryColumn = this.props.getPosition;
            if (geometryColumn !== undefined &&
                ga.vector.isPointVector(geometryColumn)) {
                return this._renderLayersPoint(geometryColumn);
            }
            throw new Error("getPosition should pass in an arrow Vector of Point type");
        }
        else {
            const pointVector = getGeometryVector(table, EXTENSION_NAME.POINT);
            if (pointVector !== null) {
                return this._renderLayersPoint(pointVector);
            }
        }
        throw new Error("getPosition not GeoArrow point");
    }
    _renderLayersPoint(geometryColumn) {
        const { data: table } = this.props;
        if (this.props._validate) {
            assert(ga.vector.isPointVector(geometryColumn));
            validateAccessors(this.props, table);
        }
        // Exclude manually-set accessors
        const [accessors, otherProps] = extractAccessorsFromProps(this.props, [
            "getPosition",
        ]);
        const tableOffsets = computeChunkOffsets(table.data);
        const layers = [];
        for (let recordBatchIdx = 0; recordBatchIdx < table.batches.length; recordBatchIdx++) {
            let geometryData = geometryColumn.data[recordBatchIdx];
            if (isGeomSeparate(geometryData)) {
                geometryData = convertStructToFixedSizeList(geometryData);
            }
            const flatCoordsData = ga.child.getPointChild(geometryData);
            const flatCoordinateArray = flatCoordsData.values;
            const props = {
                // Note: because this is a composite layer and not doing the rendering
                // itself, we still have to pass in our defaultProps
                ...ourDefaultProps$7,
                ...otherProps,
                // used for picking purposes
                recordBatchIdx,
                tableOffsets,
                id: `${this.props.id}-geoarrow-heatmap-${recordBatchIdx}`,
                data: {
                    // @ts-expect-error passed through to enable use by function accessors
                    data: table.batches[recordBatchIdx],
                    length: geometryData.length,
                    attributes: {
                        getPosition: {
                            value: flatCoordinateArray,
                            size: geometryData.type.listSize,
                        },
                    },
                },
            };
            for (const [propName, propInput] of Object.entries(accessors)) {
                assignAccessor({
                    props,
                    propName,
                    propInput,
                    chunkIdx: recordBatchIdx,
                });
            }
            const layer = new HeatmapLayer(this.getSubLayerProps(props));
            layers.push(layer);
        }
        return layers;
    }
}

// deck.gl-community
// SPDX-License-Identifier: MIT
// Copyright (c) vis.gl contributors
// Remove data and getPosition from the upstream default props
const { data: _data$6, getPath: _getPath$1, ..._defaultProps$4 } = PathLayer.defaultProps;
// Default props added by us
const ourDefaultProps$6 = {
    // Note: this diverges from upstream, where here we _default into_ binary
    // rendering
    // This instructs the layer to skip normalization and use the binary
    // as-is
    _pathType: "open",
    _validate: true,
};
// @ts-expect-error not sure why this is failing
const defaultProps$6 = {
    ..._defaultProps$4,
    ...ourDefaultProps$6,
};
/**
 * Render lists of coordinate points as extruded polylines with mitering.
 */
class GeoArrowPathLayer extends CompositeLayer {
    static defaultProps = defaultProps$6;
    static layerName = "GeoArrowPathLayer";
    getPickingInfo(params) {
        return getPickingInfo(params, this.props.data);
    }
    renderLayers() {
        const { data: table } = this.props;
        if (this.props.getPath !== undefined) {
            const geometryColumn = this.props.getPath;
            if (geometryColumn !== undefined &&
                ga.vector.isLineStringVector(geometryColumn)) {
                return this._renderLayersLineString(geometryColumn);
            }
            if (geometryColumn !== undefined &&
                ga.vector.isMultiLineStringVector(geometryColumn)) {
                return this._renderLayersMultiLineString(geometryColumn);
            }
            throw new Error("getPath should be an arrow Vector of LineString or MultiLineString type");
        }
        else {
            const lineStringVector = getGeometryVector(table, EXTENSION_NAME.LINESTRING);
            if (lineStringVector !== null) {
                return this._renderLayersLineString(lineStringVector);
            }
            const multiLineStringVector = getGeometryVector(table, EXTENSION_NAME.MULTILINESTRING);
            if (multiLineStringVector !== null) {
                return this._renderLayersMultiLineString(multiLineStringVector);
            }
        }
        throw new Error("getPath not GeoArrow LineString or MultiLineString");
    }
    _renderLayersLineString(geometryColumn) {
        const { data: table } = this.props;
        // TODO: validate that if nested, accessor props have the same nesting
        // structure as the main geometry column.
        if (this.props._validate) {
            assert(ga.vector.isLineStringVector(geometryColumn));
            validateAccessors(this.props, table);
        }
        // Exclude manually-set accessors
        const [accessors, otherProps] = extractAccessorsFromProps(this.props, [
            "getPath",
        ]);
        const tableOffsets = computeChunkOffsets(table.data);
        const layers = [];
        for (let recordBatchIdx = 0; recordBatchIdx < table.batches.length; recordBatchIdx++) {
            let lineStringData = geometryColumn.data[recordBatchIdx];
            if (isGeomSeparate(lineStringData)) {
                lineStringData = getInterleavedLineString(lineStringData);
            }
            const geomOffsets = lineStringData.valueOffsets;
            const pointData = ga.child.getLineStringChild(lineStringData);
            const nDim = pointData.type.listSize;
            const coordData = ga.child.getPointChild(pointData);
            const flatCoordinateArray = coordData.values;
            const props = {
                // Note: because this is a composite layer and not doing the rendering
                // itself, we still have to pass in our defaultProps
                ...ourDefaultProps$6,
                ...otherProps,
                // used for picking purposes
                recordBatchIdx,
                tableOffsets,
                id: `${this.props.id}-geoarrow-path-${recordBatchIdx}`,
                data: {
                    // @ts-expect-error passed through to enable use by function accessors
                    data: table.batches[recordBatchIdx],
                    length: lineStringData.length,
                    startIndices: geomOffsets,
                    attributes: {
                        getPath: { value: flatCoordinateArray, size: nDim },
                    },
                },
            };
            for (const [propName, propInput] of Object.entries(accessors)) {
                assignAccessor({
                    props,
                    propName,
                    propInput,
                    chunkIdx: recordBatchIdx,
                    geomCoordOffsets: geomOffsets,
                });
            }
            const layer = new PathLayer(this.getSubLayerProps(props));
            layers.push(layer);
        }
        return layers;
    }
    _renderLayersMultiLineString(geometryColumn) {
        const { data: table } = this.props;
        // TODO: validate that if nested, accessor props have the same nesting
        // structure as the main geometry column.
        if (this.props._validate) {
            assert(ga.vector.isMultiLineStringVector(geometryColumn));
            validateAccessors(this.props, table);
        }
        // Exclude manually-set accessors
        const [accessors, otherProps] = extractAccessorsFromProps(this.props, [
            "getPath",
        ]);
        const tableOffsets = computeChunkOffsets(table.data);
        const layers = [];
        for (let recordBatchIdx = 0; recordBatchIdx < table.batches.length; recordBatchIdx++) {
            const multiLineStringData = geometryColumn.data[recordBatchIdx];
            let lineStringData = ga.child.getMultiLineStringChild(multiLineStringData);
            if (isGeomSeparate(lineStringData)) {
                lineStringData = getInterleavedLineString(lineStringData);
            }
            const pointData = ga.child.getLineStringChild(lineStringData);
            const coordData = ga.child.getPointChild(pointData);
            const geomOffsets = multiLineStringData.valueOffsets;
            const ringOffsets = lineStringData.valueOffsets;
            const nDim = pointData.type.listSize;
            const flatCoordinateArray = coordData.values;
            const multiLineStringToCoordOffsets = getMultiLineStringResolvedOffsets(multiLineStringData);
            const props = {
                // Note: because this is a composite layer and not doing the rendering
                // itself, we still have to pass in our defaultProps
                ...ourDefaultProps$6,
                ...otherProps,
                // used for picking purposes
                recordBatchIdx,
                tableOffsets,
                id: `${this.props.id}-geoarrow-path-${recordBatchIdx}`,
                data: {
                    // @ts-expect-error passed through to enable use by function accessors
                    data: table.batches[recordBatchIdx],
                    // Map from expanded multi-geometry index to original index
                    // Used both in picking and for function callbacks
                    invertedGeomOffsets: invertOffsets(geomOffsets),
                    // Note: this needs to be the length one level down.
                    length: lineStringData.length,
                    // Offsets into coordinateArray where each single-line string starts
                    //
                    // Note: this is ringOffsets, not geomOffsets because we're rendering
                    // the individual paths on the map.
                    startIndices: ringOffsets,
                    attributes: {
                        getPath: { value: flatCoordinateArray, size: nDim },
                    },
                },
            };
            // Note: here we use multiLineStringToCoordOffsets, not ringOffsets,
            // because we want the mapping from the _feature_ to the vertex
            for (const [propName, propInput] of Object.entries(accessors)) {
                assignAccessor({
                    props,
                    propName,
                    propInput,
                    chunkIdx: recordBatchIdx,
                    geomCoordOffsets: multiLineStringToCoordOffsets,
                });
            }
            const layer = new PathLayer(this.getSubLayerProps(props));
            layers.push(layer);
        }
        return layers;
    }
}

// deck.gl-community
// SPDX-License-Identifier: MIT
// Copyright (c) vis.gl contributors
// Remove data nd get Position from the upstream default props
const { data: _data$5, getPosition: _getPosition$2, ..._upstreamDefaultProps$1 } = PointCloudLayer.defaultProps;
// Default props added by us
const ourDefaultProps$5 = {
    _validate: true,
};
// @ts-expect-error Type error in merging default props with ours
const defaultProps$5 = {
    ..._upstreamDefaultProps$1,
    ...ourDefaultProps$5,
};
class GeoArrowPointCloudLayer extends CompositeLayer {
    static defaultProps = defaultProps$5;
    static layerName = "GeoArrowPointCloudLayer";
    getPickingInfo(params) {
        return getPickingInfo(params, this.props.data);
    }
    renderLayers() {
        const { data: table } = this.props;
        if (this.props.getPosition !== undefined) {
            const geometryColumn = this.props.getPosition;
            if (geometryColumn !== undefined &&
                ga.vector.isPointVector(geometryColumn)) {
                return this._renderLayersPoint(geometryColumn);
            }
            throw new Error("getPosition should pass in an arrow Vector of Point type");
        }
        else {
            const pointVector = getGeometryVector(table, EXTENSION_NAME.POINT);
            if (pointVector !== null) {
                return this._renderLayersPoint(pointVector);
            }
        }
        throw new Error("getPosition not GeoArrow point");
    }
    _renderLayersPoint(geometryColumn) {
        const { data: table } = this.props;
        if (this.props._validate) {
            assert(ga.vector.isPointVector(geometryColumn), "The geometry column is not a valid PointVector.");
            assert(geometryColumn.type.listSize === 3, "Points of a PointCloudLayer in the geometry column must be three-dimensional.");
            validateAccessors(this.props, table);
        }
        // Exclude manually-set accessors
        const [accessors, otherProps] = extractAccessorsFromProps(this.props, [
            "getPosition",
        ]);
        const tableOffsets = computeChunkOffsets(table.data);
        const layers = [];
        for (let recordBatchIdx = 0; recordBatchIdx < table.batches.length; recordBatchIdx++) {
            let geometryData = geometryColumn.data[recordBatchIdx];
            if (isGeomSeparate(geometryData)) {
                geometryData = convertStructToFixedSizeList(geometryData);
            }
            const flatCoordsData = ga.child.getPointChild(geometryData);
            const flatCoordinateArray = flatCoordsData.values;
            const props = {
                // Note: because this is a composite layer and not doing the rendering
                // itself, we still have to pass in our defaultProps
                ...ourDefaultProps$5,
                ...otherProps,
                // used for picking purposes
                recordBatchIdx,
                tableOffsets,
                id: `${this.props.id}-geoarrow-pointcloud-${recordBatchIdx}`,
                data: {
                    // @ts-expect-error passed through to enable use by function accessors
                    data: table.batches[recordBatchIdx],
                    length: geometryData.length,
                    attributes: {
                        getPosition: {
                            value: flatCoordinateArray,
                            size: geometryData.type.listSize,
                        },
                    },
                },
            };
            for (const [propName, propInput] of Object.entries(accessors)) {
                assignAccessor({
                    props,
                    propName,
                    propInput,
                    chunkIdx: recordBatchIdx,
                });
            }
            const layer = new PointCloudLayer(this.getSubLayerProps(props));
            layers.push(layer);
        }
        return layers;
    }
}

// deck.gl-community
// SPDX-License-Identifier: MIT
// Copyright (c) vis.gl contributors
// Remove data from the upstream default props
const { data: _data$4, getPolygon: _getPolygon$1, ..._defaultProps$3 } = SolidPolygonLayer.defaultProps;
// Default props added by us
const ourDefaultProps$4 = {
    // Note: this diverges from upstream, where here we default to no
    // normalization
    _normalize: false,
    // Note: this diverges from upstream, where here we default to CCW
    _windingOrder: "CCW",
    _validate: true,
    // Note: set this to current version
    earcutWorkerUrl: "https://cdn.jsdelivr.net/npm/@geoarrow/geoarrow-js@0.3.0/dist/earcut-worker.min.js",
    earcutWorkerPoolSize: 8,
};
// @ts-expect-error Type error in merging default props with ours
const defaultProps$4 = {
    ..._defaultProps$3,
    ...ourDefaultProps$4,
};
class GeoArrowSolidPolygonLayer extends CompositeLayer {
    static defaultProps = defaultProps$4;
    static layerName = "GeoArrowSolidPolygonLayer";
    initializeState(_context) {
        this.state = {
            table: null,
            tableOffsets: null,
            triangles: null,
            earcutWorkerRequest: this.props.earcutWorkerUrl === null ||
                this.props.earcutWorkerUrl === undefined
                ? null
                : fetch(this.props.earcutWorkerUrl).then((resp) => resp.text()),
            earcutWorkerPool: null,
        };
    }
    // NOTE: I'm not 100% on the race condition implications of this; can we make
    // sure we never construct two pools?
    async initEarcutPool() {
        if (this.state.earcutWorkerPool)
            return this.state.earcutWorkerPool;
        const workerText = await this.state.earcutWorkerRequest;
        if (!workerText) {
            return null;
        }
        // Some environments are not able to execute `importScripts`
        // E.g. on a non-served HTML file (e.g. from lonboard export) you get
        // Uncaught DOMException: Failed to execute 'importScripts' on
        // 'WorkerGlobalScope': The script at
        // 'blob:null/4ffb0b98-d1bd-4d9e-be52-998f50829723' failed to load.
        //
        // Additionally, it appears impossible to _catch_ this exception (at least
        // on Chrome), so we'll hack around this by additionally checking if the
        // current file is served from file://
        if (window?.location?.href.startsWith("file://")) {
            return null;
        }
        try {
            const pool = Pool(() => spawn(BlobWorker.fromText(workerText)), 8);
            this.state.earcutWorkerPool = pool;
            return this.state.earcutWorkerPool;
        }
        catch (err) {
            return null;
        }
    }
    async finalizeState(context) {
        await this.state?.earcutWorkerPool?.terminate();
        console.log("terminated");
    }
    async updateData() {
        const { data: table } = this.props;
        const earcutTriangles = await this._updateEarcut(table);
        const tableOffsets = computeChunkOffsets(table.data);
        this.setState({
            table: this.props.data,
            triangles: earcutTriangles,
            tableOffsets,
        });
    }
    async _updateEarcut(table) {
        const polygonVector = getGeometryVector(table, EXTENSION_NAME.POLYGON);
        if (polygonVector !== null) {
            return this._earcutPolygonVector(polygonVector);
        }
        const MultiPolygonVector = getGeometryVector(table, EXTENSION_NAME.MULTIPOLYGON);
        if (MultiPolygonVector !== null) {
            return this._earcutMultiPolygonVector(MultiPolygonVector);
        }
        const geometryColumn = this.props.getPolygon;
        if (geometryColumn !== undefined &&
            ga.vector.isPolygonVector(geometryColumn)) {
            return this._earcutPolygonVector(geometryColumn);
        }
        if (geometryColumn !== undefined &&
            ga.vector.isMultiPolygonVector(geometryColumn)) {
            return this._earcutMultiPolygonVector(geometryColumn);
        }
        throw new Error("geometryColumn not Polygon or MultiPolygon");
    }
    async _earcutPolygonVector(geometryColumn) {
        const pool = await this.initEarcutPool();
        // Fallback if pool couldn't be created
        if (!pool) {
            return this._earcutPolygonVectorMainThread(geometryColumn);
        }
        const result = new Array(geometryColumn.data.length);
        console.time("earcut");
        for (let recordBatchIdx = 0; recordBatchIdx < geometryColumn.data.length; recordBatchIdx++) {
            let polygonData = geometryColumn.data[recordBatchIdx];
            // TODO: Note here that [when applicable] we do this conversion twice -
            // one for triangulation (earcut) here and the other for rendering later.
            if (isGeomSeparate(polygonData)) {
                polygonData = getInterleavedPolygon(polygonData);
            }
            const [preparedPolygonData, arrayBuffers] = ga.worker.preparePostMessage(polygonData, true);
            pool.queue(async (earcutWorker) => {
                const earcutTriangles = await earcutWorker(Transfer(preparedPolygonData, arrayBuffers));
                result[recordBatchIdx] = earcutTriangles;
            });
        }
        await pool.completed();
        console.timeEnd("earcut");
        return result;
    }
    _earcutPolygonVectorMainThread(geometryColumn) {
        const result = new Array(geometryColumn.data.length);
        for (let recordBatchIdx = 0; recordBatchIdx < geometryColumn.data.length; recordBatchIdx++) {
            let polygonData = geometryColumn.data[recordBatchIdx];
            // TODO: Note here that [when applicable] we do this conversion twice -
            // one for triangulation (earcut) here and the other for rendering later.
            if (isGeomSeparate(polygonData)) {
                polygonData = getInterleavedPolygon(polygonData);
            }
            result[recordBatchIdx] = ga.algorithm.earcut(polygonData);
        }
        return result;
    }
    async _earcutMultiPolygonVector(geometryColumn) {
        const pool = await this.initEarcutPool();
        // Fallback if pool couldn't be created
        if (!pool) {
            return this._earcutMultiPolygonVectorMainThread(geometryColumn);
        }
        const result = new Array(geometryColumn.data.length);
        console.time("earcut");
        for (let recordBatchIdx = 0; recordBatchIdx < geometryColumn.data.length; recordBatchIdx++) {
            const multiPolygonData = geometryColumn.data[recordBatchIdx];
            let polygonData = ga.child.getMultiPolygonChild(multiPolygonData);
            // TODO: Note here that [when applicable] we do this conversion twice -
            // one for triangulation (earcut) here and the other for rendering later.
            if (isGeomSeparate(polygonData)) {
                polygonData = getInterleavedPolygon(polygonData);
            }
            const [preparedPolygonData, arrayBuffers] = ga.worker.preparePostMessage(polygonData, true);
            pool.queue(async (earcutWorker) => {
                const earcutTriangles = await earcutWorker(Transfer(preparedPolygonData, arrayBuffers));
                result[recordBatchIdx] = earcutTriangles;
            });
        }
        await pool.completed();
        console.timeEnd("earcut");
        return result;
    }
    _earcutMultiPolygonVectorMainThread(geometryColumn) {
        const result = new Array(geometryColumn.data.length);
        for (let recordBatchIdx = 0; recordBatchIdx < geometryColumn.data.length; recordBatchIdx++) {
            const multiPolygonData = geometryColumn.data[recordBatchIdx];
            let polygonData = ga.child.getMultiPolygonChild(multiPolygonData);
            // TODO: Note here that [when applicable] we do this conversion twice -
            // one for triangulation (earcut) here and the other for rendering later.
            if (isGeomSeparate(polygonData)) {
                polygonData = getInterleavedPolygon(polygonData);
            }
            result[recordBatchIdx] = ga.algorithm.earcut(polygonData);
        }
        return result;
    }
    updateState({ props, changeFlags }) {
        if (changeFlags.dataChanged) {
            this.updateData();
        }
    }
    getPickingInfo(params) {
        return getPickingInfo(params, this.props.data);
    }
    renderLayers() {
        const { table } = this.state;
        if (!table)
            return null;
        if (this.props.getPolygon !== undefined) {
            const geometryColumn = this.props.getPolygon;
            if (geometryColumn !== undefined &&
                ga.vector.isPolygonVector(geometryColumn)) {
                return this._renderLayersPolygon(geometryColumn);
            }
            if (geometryColumn !== undefined &&
                ga.vector.isMultiPolygonVector(geometryColumn)) {
                return this._renderLayersMultiPolygon(geometryColumn);
            }
            throw new Error("getPolygon should be an arrow Vector of Polygon or MultiPolygon type");
        }
        else {
            const polygonVector = getGeometryVector(table, EXTENSION_NAME.POLYGON);
            if (polygonVector !== null) {
                return this._renderLayersPolygon(polygonVector);
            }
            const MultiPolygonVector = getGeometryVector(table, EXTENSION_NAME.MULTIPOLYGON);
            if (MultiPolygonVector !== null) {
                return this._renderLayersMultiPolygon(MultiPolygonVector);
            }
        }
        throw new Error("getPolygon not GeoArrow Polygon or MultiPolygon");
    }
    _renderLayersPolygon(geometryColumn) {
        const { table } = this.state;
        if (!table)
            return null;
        if (this.props._validate) {
            assert(ga.vector.isPolygonVector(geometryColumn));
            validateAccessors(this.props, table);
        }
        // Exclude manually-set accessors
        const [accessors, otherProps] = extractAccessorsFromProps(this.props, [
            "getPolygon",
        ]);
        const layers = [];
        for (let recordBatchIdx = 0; recordBatchIdx < table.batches.length; recordBatchIdx++) {
            let polygonData = geometryColumn.data[recordBatchIdx];
            if (isGeomSeparate(polygonData)) {
                polygonData = getInterleavedPolygon(polygonData);
            }
            const ringData = ga.child.getPolygonChild(polygonData);
            const pointData = ga.child.getLineStringChild(ringData);
            const coordData = ga.child.getPointChild(pointData);
            const nDim = pointData.type.listSize;
            const flatCoordinateArray = coordData.values;
            const resolvedRingOffsets = getPolygonResolvedOffsets(polygonData);
            if (!this.state.triangles) {
                return null;
            }
            const earcutTriangles = this.state.triangles[recordBatchIdx];
            const props = {
                // Note: because this is a composite layer and not doing the rendering
                // itself, we still have to pass in our defaultProps
                ...ourDefaultProps$4,
                ...otherProps,
                // used for picking purposes
                recordBatchIdx,
                tableOffsets: this.state.tableOffsets,
                id: `${this.props.id}-geoarrow-point-${recordBatchIdx}`,
                data: {
                    // @ts-expect-error passed through to enable use by function accessors
                    data: table.batches[recordBatchIdx],
                    // Number of geometries
                    length: polygonData.length,
                    // Offsets into coordinateArray where each polygon starts
                    startIndices: resolvedRingOffsets,
                    attributes: {
                        getPolygon: { value: flatCoordinateArray, size: nDim },
                        indices: { value: earcutTriangles, size: 1 },
                    },
                },
            };
            for (const [propName, propInput] of Object.entries(accessors)) {
                assignAccessor({
                    props,
                    propName,
                    propInput,
                    chunkIdx: recordBatchIdx,
                    geomCoordOffsets: resolvedRingOffsets,
                });
            }
            const layer = new SolidPolygonLayer(this.getSubLayerProps(props));
            layers.push(layer);
        }
        return layers;
    }
    _renderLayersMultiPolygon(geometryColumn) {
        const { table } = this.state;
        if (!table)
            return null;
        if (this.props._validate) {
            assert(ga.vector.isMultiPolygonVector(geometryColumn));
            validateAccessors(this.props, table);
        }
        // Exclude manually-set accessors
        const [accessors, otherProps] = extractAccessorsFromProps(this.props, [
            "getPolygon",
        ]);
        const layers = [];
        for (let recordBatchIdx = 0; recordBatchIdx < table.batches.length; recordBatchIdx++) {
            const multiPolygonData = geometryColumn.data[recordBatchIdx];
            let polygonData = ga.child.getMultiPolygonChild(multiPolygonData);
            if (isGeomSeparate(polygonData)) {
                polygonData = getInterleavedPolygon(polygonData);
            }
            const ringData = ga.child.getPolygonChild(polygonData);
            const pointData = ga.child.getLineStringChild(ringData);
            const coordData = ga.child.getPointChild(pointData);
            const nDim = pointData.type.listSize;
            const geomOffsets = multiPolygonData.valueOffsets;
            const flatCoordinateArray = coordData.values;
            if (!this.state.triangles) {
                return null;
            }
            const earcutTriangles = this.state.triangles[recordBatchIdx];
            // NOTE: we have two different uses of offsets. One is for _rendering_
            // each polygon. The other is for mapping _accessor attributes_ from one
            // value per feature to one value per vertex. And for that we need to use
            // these offsets in two different ways.
            //
            // TODO: Don't construct the offsets twice from scratch? I.e. from the
            // polygon-to-coord offsets you should be able to infer the
            // multi-polygon-to-coord offsets? Or something like that
            const resolvedPolygonToCoordOffsets = getPolygonResolvedOffsets(polygonData);
            const resolvedMultiPolygonToCoordOffsets = getMultiPolygonResolvedOffsets(multiPolygonData);
            const props = {
                // Note: because this is a composite layer and not doing the rendering
                // itself, we still have to pass in our defaultProps
                ...ourDefaultProps$4,
                ...otherProps,
                // used for picking purposes
                recordBatchIdx,
                tableOffsets: this.state.tableOffsets,
                id: `${this.props.id}-geoarrow-solid-polygon-multi-${recordBatchIdx}`,
                data: {
                    // @ts-expect-error passed through to enable use by function accessors
                    data: table.batches[recordBatchIdx],
                    // Map from expanded multi-geometry index to original index
                    // Used both in picking and for function callbacks
                    invertedGeomOffsets: invertOffsets(geomOffsets),
                    // Number of polygons
                    // Note: this needs to be the length one level down, because we're
                    // rendering the polygons, not the multipolygons
                    length: polygonData.length,
                    // Offsets into coordinateArray where each single-polygon starts
                    //
                    // Note that this is polygonToCoordOffsets and not geomToCoordOffsets
                    // because we're rendering each part of the MultiPolygon individually
                    startIndices: resolvedPolygonToCoordOffsets,
                    attributes: {
                        getPolygon: { value: flatCoordinateArray, size: nDim },
                        indices: { value: earcutTriangles, size: 1 },
                        // instancePickingColors: {
                        //   value: encodePickingColors(
                        //     resolvedMultiPolygonToCoordOffsets,
                        //     this.encodePickingColor
                        //   ),
                        //   size: 3,
                        // },
                    },
                },
            };
            for (const [propName, propInput] of Object.entries(accessors)) {
                assignAccessor({
                    props,
                    propName,
                    propInput,
                    chunkIdx: recordBatchIdx,
                    geomCoordOffsets: resolvedMultiPolygonToCoordOffsets,
                });
            }
            const finalProps = this.getSubLayerProps(props);
            const layer = new SolidPolygonLayer(finalProps);
            layers.push(layer);
        }
        return layers;
    }
}

// deck.gl-community
// SPDX-License-Identifier: MIT
// Copyright (c) vis.gl contributors
function getPolygonExterior(input) {
    if ("data" in input) {
        return new arrow.Vector(input.data.map((data) => getPolygonExterior(data)));
    }
    return input;
}
function getMultiPolygonExterior(input) {
    if ("data" in input) {
        return new arrow.Vector(input.data.map((data) => getMultiPolygonExterior(data)));
    }
    const geomOffsets = input.valueOffsets;
    const polygonData = ga.child.getMultiPolygonChild(input);
    const polygonOffsets = polygonData.valueOffsets;
    const lineStringData = ga.child.getPolygonChild(polygonData);
    const resolvedOffsets = new Int32Array(geomOffsets.length);
    for (let i = 0; i < resolvedOffsets.length; ++i) {
        // Perform the lookup
        resolvedOffsets[i] = polygonOffsets[geomOffsets[i]];
    }
    return arrow.makeData({
        type: new arrow.List(polygonData.type.children[0]),
        length: input.length,
        nullCount: input.nullCount,
        nullBitmap: input.nullBitmap,
        child: lineStringData,
        valueOffsets: resolvedOffsets,
    });
}
// Remove data and getPolygon from the upstream default props
const { data: _data$3, getPolygon: _getPolygon, ..._defaultProps$2 } = PolygonLayer.defaultProps;
// Default props added by us
const ourDefaultProps$3 = {
    // Note: this diverges from upstream, where here we default to no
    // normalization
    _normalize: false,
    // Note: this diverges from upstream, where here we default to CCW
    _windingOrder: "CCW",
    _validate: true,
};
// @ts-expect-error Type error in merging default props with ours
const defaultProps$3 = {
    ..._defaultProps$2,
    ...ourDefaultProps$3,
};
const defaultLineColor = [0, 0, 0, 255];
/** The `GeoArrowPolygonLayer` renders filled, stroked and/or extruded polygons.
 *
 * GeoArrowPolygonLayer is a CompositeLayer that wraps the
 * GeoArrowSolidPolygonLayer and the GeoArrowPathLayer.
 */
class GeoArrowPolygonLayer extends CompositeLayer {
    static defaultProps = defaultProps$3;
    static layerName = "GeoArrowPolygonLayer";
    getPickingInfo(params) {
        // Propagate the picked info from the SolidPolygonLayer
        return params.info;
    }
    renderLayers() {
        const { data: table } = this.props;
        if (this.props.getPolygon !== undefined) {
            const geometryColumn = this.props.getPolygon;
            if (ga.vector.isPolygonVector(geometryColumn)) {
                return this._renderLayers(geometryColumn);
            }
            if (ga.vector.isMultiPolygonVector(geometryColumn)) {
                return this._renderLayers(geometryColumn);
            }
            throw new Error("getPolygon should be an arrow Vector of Polygon or MultiPolygon type");
        }
        else {
            const polygonVector = getGeometryVector(table, EXTENSION_NAME.POLYGON);
            if (polygonVector !== null) {
                return this._renderLayers(polygonVector);
            }
            const multiPolygonVector = getGeometryVector(table, EXTENSION_NAME.MULTIPOLYGON);
            if (multiPolygonVector !== null) {
                return this._renderLayers(multiPolygonVector);
            }
        }
        throw new Error("geometryColumn not Polygon or MultiPolygon");
    }
    // NOTE: Here we shouldn't need a split for handling both multi- and single-
    // geometries, because the underlying SolidPolygonLayer and PathLayer both
    // support multi-* and single- geometries.
    _renderLayers(geometryColumn) {
        const { data: table } = this.props;
        let getPath;
        if (ga.vector.isPolygonVector(geometryColumn)) {
            getPath = getPolygonExterior(geometryColumn);
        }
        else if (ga.vector.isMultiPolygonVector(geometryColumn)) {
            getPath = getMultiPolygonExterior(geometryColumn);
        }
        else {
            assert(false);
        }
        // Layer composition props
        const { data, _dataDiff, stroked, filled, extruded, wireframe, _normalize, _windingOrder, elevationScale, transitions, positionFormat, } = this.props;
        // Rendering props underlying layer
        const { lineWidthUnits, lineWidthScale, lineWidthMinPixels, lineWidthMaxPixels, lineJointRounded, lineMiterLimit, lineDashJustified, } = this.props;
        // Accessor props for underlying layers
        const { getFillColor, getLineColor, getLineWidth, getElevation, getPolygon, updateTriggers, material, } = this.props;
        const FillLayer = this.getSubLayerClass("fill", GeoArrowSolidPolygonLayer);
        const StrokeLayer = this.getSubLayerClass("stroke", GeoArrowPathLayer);
        // Filled Polygon Layer
        const polygonLayer = new FillLayer({
            // _dataDiff,
            extruded,
            elevationScale,
            filled,
            wireframe,
            _normalize,
            _windingOrder,
            getElevation,
            getFillColor,
            getLineColor: extruded && wireframe ? getLineColor : defaultLineColor,
            material,
            transitions,
        }, this.getSubLayerProps({
            id: "fill",
            updateTriggers: updateTriggers && {
                getPolygon: updateTriggers.getPolygon,
                getElevation: updateTriggers.getElevation,
                getFillColor: updateTriggers.getFillColor,
                getLineColor: updateTriggers.getLineColor,
            },
        }), {
            data,
            positionFormat,
            getPolygon,
        });
        // Polygon line layer
        const polygonLineLayer = !extruded &&
            stroked &&
            new StrokeLayer({
                // _dataDiff,
                widthUnits: lineWidthUnits,
                widthScale: lineWidthScale,
                widthMinPixels: lineWidthMinPixels,
                widthMaxPixels: lineWidthMaxPixels,
                jointRounded: lineJointRounded,
                miterLimit: lineMiterLimit,
                dashJustified: lineDashJustified,
                // Already normalized, and since they had been polygons, we know that
                // the lines are a loop.
                _pathType: "loop",
                transitions: transitions && {
                    getWidth: transitions.getLineWidth,
                    getColor: transitions.getLineColor,
                    getPath: transitions.getPolygon,
                },
                getColor: this.getSubLayerAccessor(getLineColor),
                getWidth: this.getSubLayerAccessor(getLineWidth),
            }, this.getSubLayerProps({
                id: "stroke",
                updateTriggers: updateTriggers && {
                    getWidth: updateTriggers.getLineWidth,
                    getColor: updateTriggers.getLineColor,
                    getDashArray: updateTriggers.getLineDashArray,
                },
            }), {
                data: table,
                positionFormat,
                getPath,
                // We only pick solid polygon layers, not the path layers
                pickable: false,
            });
        const layers = [
            // If not extruded: flat fill layer is drawn below outlines
            !extruded && polygonLayer,
            polygonLineLayer,
            // If extruded: draw fill layer last for correct blending behavior
            extruded && polygonLayer,
        ];
        return layers;
    }
}

// deck.gl-community
// SPDX-License-Identifier: MIT
// Copyright (c) vis.gl contributors
// Remove data and getPosition from the upstream default props
const { data: _data$2, getPosition: _getPosition$1, ..._upstreamDefaultProps } = ScatterplotLayer.defaultProps;
// Default props added by us
const ourDefaultProps$2 = {
    _validate: true,
};
// @ts-expect-error Type error in merging default props with ours
const defaultProps$2 = {
    ..._upstreamDefaultProps,
    ...ourDefaultProps$2,
};
class GeoArrowScatterplotLayer extends CompositeLayer {
    static defaultProps = defaultProps$2;
    static layerName = "GeoArrowScatterplotLayer";
    getPickingInfo(params) {
        return getPickingInfo(params, this.props.data);
    }
    renderLayers() {
        const { data: table } = this.props;
        if (this.props.getPosition !== undefined) {
            const geometryColumn = this.props.getPosition;
            if (geometryColumn !== undefined &&
                ga.vector.isPointVector(geometryColumn)) {
                return this._renderLayersPoint(geometryColumn);
            }
            if (geometryColumn !== undefined &&
                ga.vector.isMultiPointVector(geometryColumn)) {
                return this._renderLayersMultiPoint(geometryColumn);
            }
            throw new Error("getPosition should pass in an arrow Vector of Point or MultiPoint type");
        }
        else {
            const pointVector = getGeometryVector(table, EXTENSION_NAME.POINT);
            if (pointVector !== null) {
                return this._renderLayersPoint(pointVector);
            }
            const multiPointVector = getGeometryVector(table, EXTENSION_NAME.MULTIPOINT);
            if (multiPointVector !== null) {
                return this._renderLayersMultiPoint(multiPointVector);
            }
        }
        throw new Error("getPosition not GeoArrow point or multipoint");
    }
    _renderLayersPoint(geometryColumn) {
        const { data: table } = this.props;
        if (this.props._validate) {
            assert(ga.vector.isPointVector(geometryColumn));
            validateAccessors(this.props, table);
        }
        // Exclude manually-set accessors
        const [accessors, otherProps] = extractAccessorsFromProps(this.props, [
            "getPosition",
        ]);
        const tableOffsets = computeChunkOffsets(table.data);
        const layers = [];
        for (let recordBatchIdx = 0; recordBatchIdx < table.batches.length; recordBatchIdx++) {
            let geometryData = geometryColumn.data[recordBatchIdx];
            if (isGeomSeparate(geometryData)) {
                geometryData = convertStructToFixedSizeList(geometryData);
            }
            const flatCoordsData = ga.child.getPointChild(geometryData);
            const flatCoordinateArray = flatCoordsData.values;
            const props = {
                // Note: because this is a composite layer and not doing the rendering
                // itself, we still have to pass in our defaultProps
                ...ourDefaultProps$2,
                ...otherProps,
                // used for picking purposes
                recordBatchIdx,
                tableOffsets,
                id: `${this.props.id}-geoarrow-scatterplot-${recordBatchIdx}`,
                data: {
                    // @ts-expect-error passed through to enable use by function accessors
                    data: table.batches[recordBatchIdx],
                    length: geometryData.length,
                    attributes: {
                        getPosition: {
                            value: flatCoordinateArray,
                            size: geometryData.type.listSize,
                        },
                    },
                },
            };
            for (const [propName, propInput] of Object.entries(accessors)) {
                assignAccessor({
                    props,
                    propName,
                    propInput,
                    chunkIdx: recordBatchIdx,
                });
            }
            const layer = new ScatterplotLayer(this.getSubLayerProps(props));
            layers.push(layer);
        }
        return layers;
    }
    _renderLayersMultiPoint(geometryColumn) {
        const { data: table } = this.props;
        // TODO: validate that if nested, accessor props have the same nesting
        // structure as the main geometry column.
        if (this.props._validate) {
            assert(ga.vector.isMultiPointVector(geometryColumn));
            validateAccessors(this.props, table);
        }
        // Exclude manually-set accessors
        const [accessors, otherProps] = extractAccessorsFromProps(this.props, [
            "getPosition",
        ]);
        const tableOffsets = computeChunkOffsets(table.data);
        const layers = [];
        for (let recordBatchIdx = 0; recordBatchIdx < table.batches.length; recordBatchIdx++) {
            const multiPointData = geometryColumn.data[recordBatchIdx];
            let pointData = ga.child.getMultiPointChild(multiPointData);
            if (isGeomSeparate(pointData)) {
                pointData = convertStructToFixedSizeList(pointData);
            }
            const geomOffsets = multiPointData.valueOffsets;
            const flatCoordsData = ga.child.getPointChild(pointData);
            const flatCoordinateArray = flatCoordsData.values;
            const props = {
                // Note: because this is a composite layer and not doing the rendering
                // itself, we still have to pass in our defaultProps
                ...ourDefaultProps$2,
                ...otherProps,
                // used for picking purposes
                recordBatchIdx,
                tableOffsets,
                id: `${this.props.id}-geoarrow-scatterplot-${recordBatchIdx}`,
                data: {
                    // @ts-expect-error passed through to enable use by function accessors
                    data: table.batches[recordBatchIdx],
                    // Map from expanded multi-geometry index to original index
                    // Used both in picking and for function callbacks
                    invertedGeomOffsets: invertOffsets(geomOffsets),
                    // Note: this needs to be the length one level down.
                    length: pointData.length,
                    attributes: {
                        getPosition: {
                            value: flatCoordinateArray,
                            size: pointData.type.listSize,
                        },
                    },
                },
            };
            for (const [propName, propInput] of Object.entries(accessors)) {
                assignAccessor({
                    props,
                    propName,
                    propInput,
                    chunkIdx: recordBatchIdx,
                    geomCoordOffsets: geomOffsets,
                });
            }
            const layer = new ScatterplotLayer(this.getSubLayerProps(props));
            layers.push(layer);
        }
        return layers;
    }
}

// deck.gl-community
// SPDX-License-Identifier: MIT
// Copyright (c) vis.gl contributors
// Remove data and getPosition from the upstream default props
const { data: _data$1, getPosition: _getPosition, getText: _getText, getTextAnchor: _getTextAnchor, getAlignmentBaseline: _getAlignmentBaseline, getPixelOffset: _getPixelOffset, ..._defaultProps$1 } = TextLayer.defaultProps;
// Default props added by us
const ourDefaultProps$1 = {
    getTextAnchor: "middle",
    getAlignmentBaseline: "center",
    getPixelOffset: [0, 0],
    _validate: true,
};
// @ts-expect-error Type error in merging default props with ours
const defaultProps$1 = {
    ..._defaultProps$1,
    ...ourDefaultProps$1,
};
class GeoArrowTextLayer extends CompositeLayer {
    static defaultProps = defaultProps$1;
    static layerName = "GeoArrowTextLayer";
    getPickingInfo(params) {
        return getPickingInfo(params, this.props.data);
    }
    renderLayers() {
        const { data: table } = this.props;
        if (this.props.getPosition !== undefined) {
            const geometryColumn = this.props.getPosition;
            if (geometryColumn !== undefined &&
                ga.vector.isPointVector(geometryColumn)) {
                return this._renderLayersPoint(geometryColumn);
            }
            throw new Error("getPosition should pass in an arrow Vector of Point type");
        }
        else {
            const pointVector = getGeometryVector(table, EXTENSION_NAME.POINT);
            if (pointVector !== null) {
                return this._renderLayersPoint(pointVector);
            }
        }
        throw new Error("getPosition not GeoArrow point");
    }
    _renderLayersPoint(geometryColumn) {
        const { data: table } = this.props;
        if (this.props._validate) {
            assert(ga.vector.isPointVector(geometryColumn));
            validateAccessors(this.props, table);
        }
        // Exclude manually-set accessors
        const [accessors, otherProps] = extractAccessorsFromProps(this.props, [
            "getPosition",
            "getText",
        ]);
        const tableOffsets = computeChunkOffsets(table.data);
        const layers = [];
        for (let recordBatchIdx = 0; recordBatchIdx < table.batches.length; recordBatchIdx++) {
            let geometryData = geometryColumn.data[recordBatchIdx];
            if (isGeomSeparate(geometryData)) {
                geometryData = convertStructToFixedSizeList(geometryData);
            }
            const flatCoordsData = ga.child.getPointChild(geometryData);
            const flatCoordinateArray = flatCoordsData.values;
            const textData = this.props.getText.data[recordBatchIdx];
            // console.log(textData);
            const textValues = textData.values;
            const characterOffsets = textData.valueOffsets;
            const props = {
                // Note: because this is a composite layer and not doing the rendering
                // itself, we still have to pass in our defaultProps
                ...ourDefaultProps$1,
                ...otherProps,
                // used for picking purposes
                recordBatchIdx,
                tableOffsets,
                id: `${this.props.id}-geoarrow-heatmap-${recordBatchIdx}`,
                data: {
                    // @ts-expect-error passed through to enable use by function accessors
                    data: table.batches[recordBatchIdx],
                    length: geometryData.length,
                    startIndices: characterOffsets,
                    attributes: {
                        // Positions need to be expanded to be one per character!
                        getPosition: {
                            value: expandArrayToCoords(flatCoordinateArray, geometryData.type.listSize, characterOffsets),
                            size: geometryData.type.listSize,
                        },
                        // TODO: support non-ascii characters
                        getText: {
                            value: textValues,
                            // size: 1,
                        },
                    },
                },
            };
            for (const [propName, propInput] of Object.entries(accessors)) {
                assignAccessor({
                    props,
                    propName,
                    propInput,
                    chunkIdx: recordBatchIdx,
                    geomCoordOffsets: characterOffsets,
                });
            }
            const layer = new TextLayer(this.getSubLayerProps(props));
            layers.push(layer);
        }
        return layers;
    }
}

// deck.gl-community
// SPDX-License-Identifier: MIT
// Copyright (c) vis.gl contributors
// Remove data and getPosition from the upstream default props
const { data: _data, getPath: _getPath, ..._defaultProps } = defaultProps$6;
// Default props added by us
const ourDefaultProps = {
    // Note: this diverges from upstream, where here we _default into_ binary
    // rendering
    // This instructs the layer to skip normalization and use the binary
    // as-is
    _pathType: "open",
    _validate: true,
};
const defaultProps = {
    ..._defaultProps,
    ...ourDefaultProps,
};
/** Render animated paths that represent vehicle trips. */
class GeoArrowTripsLayer extends CompositeLayer {
    static defaultProps = defaultProps;
    static layerName = "GeoArrowTripsLayer";
    renderLayers() {
        const { data: table } = this.props;
        if (this.props.getPath !== undefined) {
            const geometryColumn = this.props.getPath;
            if (geometryColumn !== undefined &&
                ga.vector.isLineStringVector(geometryColumn)) {
                return this._renderLayersLineString(geometryColumn);
            }
            throw new Error("getPath should be an arrow Vector of LineString type");
        }
        else {
            const lineStringVector = getGeometryVector(table, EXTENSION_NAME.LINESTRING);
            if (lineStringVector !== null) {
                return this._renderLayersLineString(lineStringVector);
            }
        }
        throw new Error("getPath not GeoArrow LineString");
    }
    _renderLayersLineString(geometryColumn) {
        const { data: table } = this.props;
        const timestampColumn = this.props.getTimestamps;
        if (this.props._validate) {
            assert(ga.vector.isLineStringVector(geometryColumn));
            validateAccessors(this.props, table);
        }
        // Exclude manually-set accessors
        const [accessors, otherProps] = extractAccessorsFromProps(this.props, [
            "getPath",
            "getTimestamps",
        ]);
        const tableOffsets = computeChunkOffsets(table.data);
        const layers = [];
        for (let recordBatchIdx = 0; recordBatchIdx < table.batches.length; recordBatchIdx++) {
            let lineStringData = geometryColumn.data[recordBatchIdx];
            if (isGeomSeparate(lineStringData)) {
                lineStringData = getInterleavedLineString(lineStringData);
            }
            const geomOffsets = lineStringData.valueOffsets;
            const pointData = ga.child.getLineStringChild(lineStringData);
            const nDim = pointData.type.listSize;
            const coordData = ga.child.getPointChild(pointData);
            const flatCoordinateArray = coordData.values;
            const timestampData = timestampColumn.data[recordBatchIdx];
            const timestampValues = timestampData.children[0].values;
            const props = {
                // Note: because this is a composite layer and not doing the rendering
                // itself, we still have to pass in our defaultProps
                ...ourDefaultProps,
                ...otherProps,
                // used for picking purposes
                recordBatchIdx,
                tableOffsets,
                id: `${this.props.id}-geoarrow-trip-${recordBatchIdx}`,
                data: {
                    // @ts-expect-error passed through to enable use by function accessors
                    data: table.batches[recordBatchIdx],
                    length: lineStringData.length,
                    startIndices: geomOffsets,
                    attributes: {
                        getPath: { value: flatCoordinateArray, size: nDim },
                        getTimestamps: { value: timestampValues, size: 1 },
                    },
                },
            };
            for (const [propName, propInput] of Object.entries(accessors)) {
                assignAccessor({
                    props,
                    propName,
                    propInput,
                    chunkIdx: recordBatchIdx,
                    geomCoordOffsets: geomOffsets,
                });
            }
            const layer = new TripsLayer(props);
            layers.push(layer);
        }
        return layers;
    }
}

export { GeoArrowArcLayer, GeoArrowColumnLayer, GeoArrowHeatmapLayer, GeoArrowPathLayer, GeoArrowPointCloudLayer, GeoArrowPolygonLayer, GeoArrowScatterplotLayer, GeoArrowSolidPolygonLayer, GeoArrowTripsLayer, GeoArrowH3HexagonLayer as _GeoArrowH3HexagonLayer, GeoArrowTextLayer as _GeoArrowTextLayer };
//# sourceMappingURL=dist.es.mjs.map
