// deck.gl-community
// SPDX-License-Identifier: MIT
// Copyright (c) vis.gl contributors
import { CompositeLayer, assert, } from "@deck.gl/core";
import { ArcLayer } from "@deck.gl/layers";
import * as ga from "@geoarrow/geoarrow-js";
import { assignAccessor, convertStructToFixedSizeList, extractAccessorsFromProps, isGeomSeparate, } from "../utils/utils";
import { child } from "@geoarrow/geoarrow-js";
import { computeChunkOffsets, getPickingInfo, } from "../utils/picking";
import { validateAccessors } from "../utils/validate";
// Remove data from the upstream default props
const { data: _data, getSourcePosition: _getSourcePosition, getTargetPosition: _getTargetPosition, ..._defaultProps } = ArcLayer.defaultProps;
// Default props added by us
const ourDefaultProps = {
    _validate: true,
};
// @ts-expect-error Type error in merging default props with ours
const defaultProps = {
    ..._defaultProps,
    ...ourDefaultProps,
};
export class GeoArrowArcLayer extends CompositeLayer {
    static defaultProps = defaultProps;
    static layerName = "GeoArrowArcLayer";
    getPickingInfo(params) {
        return getPickingInfo(params, this.props.data);
    }
    renderLayers() {
        return this._renderLayersPoint();
    }
    _renderLayersPoint() {
        const { data: table, getSourcePosition: sourcePosition, getTargetPosition: targetPosition, } = this.props;
        if (this.props._validate) {
            validateAccessors(this.props, table);
            // Note: below we iterate over table batches anyways, so this layer won't
            // work as-is if data/table is null
            assert(ga.vector.isPointVector(sourcePosition));
            assert(ga.vector.isPointVector(targetPosition));
        }
        // Exclude manually-set accessors
        const [accessors, otherProps] = extractAccessorsFromProps(this.props, [
            "getSourcePosition",
            "getTargetPosition",
        ]);
        const tableOffsets = computeChunkOffsets(table.data);
        const layers = [];
        for (let recordBatchIdx = 0; recordBatchIdx < table.batches.length; recordBatchIdx++) {
            let sourceData = sourcePosition.data[recordBatchIdx];
            if (isGeomSeparate(sourceData)) {
                sourceData = convertStructToFixedSizeList(sourceData);
            }
            const sourceValues = child.getPointChild(sourceData).values;
            let targetData = targetPosition.data[recordBatchIdx];
            if (isGeomSeparate(targetData)) {
                targetData = convertStructToFixedSizeList(targetData);
            }
            const targetValues = child.getPointChild(targetData).values;
            const props = {
                // Note: because this is a composite layer and not doing the rendering
                // itself, we still have to pass in our defaultProps
                ...ourDefaultProps,
                ...otherProps,
                // used for picking purposes
                recordBatchIdx,
                tableOffsets,
                id: `${this.props.id}-geoarrow-arc-${recordBatchIdx}`,
                data: {
                    // @ts-expect-error passed through to enable use by function accessors
                    data: table.batches[recordBatchIdx],
                    length: sourceData.length,
                    attributes: {
                        getSourcePosition: {
                            value: sourceValues,
                            size: sourceData.type.listSize,
                        },
                        getTargetPosition: {
                            value: targetValues,
                            size: targetData.type.listSize,
                        },
                    },
                },
            };
            for (const [propName, propInput] of Object.entries(accessors)) {
                assignAccessor({
                    props,
                    propName,
                    propInput,
                    chunkIdx: recordBatchIdx,
                });
            }
            const layer = new ArcLayer(this.getSubLayerProps(props));
            layers.push(layer);
        }
        return layers;
    }
}
