// deck.gl-community
// SPDX-License-Identifier: MIT
// Copyright (c) vis.gl contributors
import { CompositeLayer, assert, } from "@deck.gl/core";
import { PointCloudLayer } from "@deck.gl/layers";
import * as ga from "@geoarrow/geoarrow-js";
import { assignAccessor, convertStructToFixedSizeList, extractAccessorsFromProps, getGeometryVector, isGeomSeparate, } from "../utils/utils";
import { computeChunkOffsets, getPickingInfo, } from "../utils/picking";
import { EXTENSION_NAME } from "../constants";
import { validateAccessors } from "../utils/validate";
// Remove data nd get Position from the upstream default props
const { data: _data, getPosition: _getPosition, ..._upstreamDefaultProps } = PointCloudLayer.defaultProps;
// Default props added by us
const ourDefaultProps = {
    _validate: true,
};
// @ts-expect-error Type error in merging default props with ours
const defaultProps = {
    ..._upstreamDefaultProps,
    ...ourDefaultProps,
};
export class GeoArrowPointCloudLayer extends CompositeLayer {
    static defaultProps = defaultProps;
    static layerName = "GeoArrowPointCloudLayer";
    getPickingInfo(params) {
        return getPickingInfo(params, this.props.data);
    }
    renderLayers() {
        const { data: table } = this.props;
        if (this.props.getPosition !== undefined) {
            const geometryColumn = this.props.getPosition;
            if (geometryColumn !== undefined &&
                ga.vector.isPointVector(geometryColumn)) {
                return this._renderLayersPoint(geometryColumn);
            }
            throw new Error("getPosition should pass in an arrow Vector of Point type");
        }
        else {
            const pointVector = getGeometryVector(table, EXTENSION_NAME.POINT);
            if (pointVector !== null) {
                return this._renderLayersPoint(pointVector);
            }
        }
        throw new Error("getPosition not GeoArrow point");
    }
    _renderLayersPoint(geometryColumn) {
        const { data: table } = this.props;
        if (this.props._validate) {
            assert(ga.vector.isPointVector(geometryColumn), "The geometry column is not a valid PointVector.");
            assert(geometryColumn.type.listSize === 3, "Points of a PointCloudLayer in the geometry column must be three-dimensional.");
            validateAccessors(this.props, table);
        }
        // Exclude manually-set accessors
        const [accessors, otherProps] = extractAccessorsFromProps(this.props, [
            "getPosition",
        ]);
        const tableOffsets = computeChunkOffsets(table.data);
        const layers = [];
        for (let recordBatchIdx = 0; recordBatchIdx < table.batches.length; recordBatchIdx++) {
            let geometryData = geometryColumn.data[recordBatchIdx];
            if (isGeomSeparate(geometryData)) {
                geometryData = convertStructToFixedSizeList(geometryData);
            }
            const flatCoordsData = ga.child.getPointChild(geometryData);
            const flatCoordinateArray = flatCoordsData.values;
            const props = {
                // Note: because this is a composite layer and not doing the rendering
                // itself, we still have to pass in our defaultProps
                ...ourDefaultProps,
                ...otherProps,
                // used for picking purposes
                recordBatchIdx,
                tableOffsets,
                id: `${this.props.id}-geoarrow-pointcloud-${recordBatchIdx}`,
                data: {
                    // @ts-expect-error passed through to enable use by function accessors
                    data: table.batches[recordBatchIdx],
                    length: geometryData.length,
                    attributes: {
                        getPosition: {
                            value: flatCoordinateArray,
                            size: geometryData.type.listSize,
                        },
                    },
                },
            };
            for (const [propName, propInput] of Object.entries(accessors)) {
                assignAccessor({
                    props,
                    propName,
                    propInput,
                    chunkIdx: recordBatchIdx,
                });
            }
            const layer = new PointCloudLayer(this.getSubLayerProps(props));
            layers.push(layer);
        }
        return layers;
    }
}
