import { CompositeLayer, CompositeLayerProps, DefaultProps, Layer, LayersList, GetPickingInfoParams } from "@deck.gl/core";
import type { PolygonLayerProps } from "@deck.gl/layers";
import * as arrow from "apache-arrow";
import * as ga from "@geoarrow/geoarrow-js";
import { GeoArrowExtraPickingProps } from "../utils/picking";
import { ColorAccessor, FloatAccessor, GeoArrowPickingInfo } from "../types";
/**
 * Get the exterior of a PolygonVector or PolygonData as a MultiLineString
 *
 * Note that casting to a MultiLineString is a no-op of the underlying data
 * structure. For the purposes of the PolygonLayer we don't want to cast to a
 * LineString because that would change the number of rows in the table.
 */
export declare function getPolygonExterior(input: ga.vector.PolygonVector): ga.vector.MultiLineStringVector;
export declare function getPolygonExterior(input: ga.data.PolygonData): ga.data.MultiLineStringData;
/**
 * Get the exterior of a MultiPolygonVector or MultiPolygonData
 *
 * Note that for the purposes of the PolygonLayer, we don't want to change the
 * number of rows in the table. Instead, we convert each MultiPolygon to a
 * single MultiLineString, combining all exteriors of each contained Polygon
 * into a single united MultiLineString.
 *
 * This means that we need to condense both two offset buffers from the
 * MultiPolygonVector/Data (geomOffsets and polygonOffsets) into a single
 * `geomOffsets` for the new MultiLineStringVector/Data.
 */
export declare function getMultiPolygonExterior(input: ga.vector.MultiPolygonVector): ga.vector.MultiLineStringVector;
export declare function getMultiPolygonExterior(input: ga.data.MultiPolygonData): ga.data.MultiLineStringData;
/** All properties supported by GeoArrowPolygonLayer */
export type GeoArrowPolygonLayerProps = Omit<PolygonLayerProps, "data" | "getPolygon" | "getFillColor" | "getLineColor" | "getLineWidth" | "getElevation"> & _GeoArrowPolygonLayerProps & CompositeLayerProps;
/** Properties added by GeoArrowPolygonLayer */
type _GeoArrowPolygonLayerProps = {
    data: arrow.Table;
    /** Polygon geometry accessor. */
    getPolygon?: ga.vector.PolygonVector | ga.vector.MultiPolygonVector;
    /** Fill color accessor.
     * @default [0, 0, 0, 255]
     */
    getFillColor?: ColorAccessor;
    /** Stroke color accessor.
     * @default [0, 0, 0, 255]
     */
    getLineColor?: ColorAccessor;
    /**
     * Line width value of accessor.
     * @default 1
     */
    getLineWidth?: FloatAccessor;
    /**
     * Elevation value of accessor.
     *
     * Only used if `extruded: true`.
     *
     * @default 1000
     */
    getElevation?: FloatAccessor;
    /**
     * If `true`, validate the arrays provided (e.g. chunk lengths)
     * @default true
     */
    _validate?: boolean;
};
/** The `GeoArrowPolygonLayer` renders filled, stroked and/or extruded polygons.
 *
 * GeoArrowPolygonLayer is a CompositeLayer that wraps the
 * GeoArrowSolidPolygonLayer and the GeoArrowPathLayer.
 */
export declare class GeoArrowPolygonLayer<ExtraProps extends {} = {}> extends CompositeLayer<Required<GeoArrowPolygonLayerProps> & ExtraProps> {
    static defaultProps: DefaultProps<GeoArrowPolygonLayerProps>;
    static layerName: string;
    getPickingInfo(params: GetPickingInfoParams & {
        sourceLayer: {
            props: GeoArrowExtraPickingProps;
        };
    }): GeoArrowPickingInfo;
    renderLayers(): Layer<{}> | LayersList | null;
    _renderLayers(geometryColumn: ga.vector.PolygonVector | ga.vector.MultiPolygonVector): Layer<{}> | LayersList | null;
}
export {};
