// deck.gl-community
// SPDX-License-Identifier: MIT
// Copyright (c) vis.gl contributors
import { CompositeLayer, assert, } from "@deck.gl/core";
import { PolygonLayer } from "@deck.gl/layers";
import * as arrow from "apache-arrow";
import * as ga from "@geoarrow/geoarrow-js";
import { getGeometryVector } from "../utils/utils";
import { EXTENSION_NAME } from "../constants";
import { GeoArrowSolidPolygonLayer } from "./solid-polygon-layer";
import { GeoArrowPathLayer } from "./path-layer";
export function getPolygonExterior(input) {
    if ("data" in input) {
        return new arrow.Vector(input.data.map((data) => getPolygonExterior(data)));
    }
    return input;
}
export function getMultiPolygonExterior(input) {
    if ("data" in input) {
        return new arrow.Vector(input.data.map((data) => getMultiPolygonExterior(data)));
    }
    const geomOffsets = input.valueOffsets;
    const polygonData = ga.child.getMultiPolygonChild(input);
    const polygonOffsets = polygonData.valueOffsets;
    const lineStringData = ga.child.getPolygonChild(polygonData);
    const resolvedOffsets = new Int32Array(geomOffsets.length);
    for (let i = 0; i < resolvedOffsets.length; ++i) {
        // Perform the lookup
        resolvedOffsets[i] = polygonOffsets[geomOffsets[i]];
    }
    return arrow.makeData({
        type: new arrow.List(polygonData.type.children[0]),
        length: input.length,
        nullCount: input.nullCount,
        nullBitmap: input.nullBitmap,
        child: lineStringData,
        valueOffsets: resolvedOffsets,
    });
}
// Remove data and getPolygon from the upstream default props
const { data: _data, getPolygon: _getPolygon, ..._defaultProps } = PolygonLayer.defaultProps;
// Default props added by us
const ourDefaultProps = {
    // Note: this diverges from upstream, where here we default to no
    // normalization
    _normalize: false,
    // Note: this diverges from upstream, where here we default to CCW
    _windingOrder: "CCW",
    _validate: true,
};
// @ts-expect-error Type error in merging default props with ours
const defaultProps = {
    ..._defaultProps,
    ...ourDefaultProps,
};
const defaultLineColor = [0, 0, 0, 255];
const defaultFillColor = [0, 0, 0, 255];
/** The `GeoArrowPolygonLayer` renders filled, stroked and/or extruded polygons.
 *
 * GeoArrowPolygonLayer is a CompositeLayer that wraps the
 * GeoArrowSolidPolygonLayer and the GeoArrowPathLayer.
 */
export class GeoArrowPolygonLayer extends CompositeLayer {
    static defaultProps = defaultProps;
    static layerName = "GeoArrowPolygonLayer";
    getPickingInfo(params) {
        // Propagate the picked info from the SolidPolygonLayer
        return params.info;
    }
    renderLayers() {
        const { data: table } = this.props;
        if (this.props.getPolygon !== undefined) {
            const geometryColumn = this.props.getPolygon;
            if (ga.vector.isPolygonVector(geometryColumn)) {
                return this._renderLayers(geometryColumn);
            }
            if (ga.vector.isMultiPolygonVector(geometryColumn)) {
                return this._renderLayers(geometryColumn);
            }
            throw new Error("getPolygon should be an arrow Vector of Polygon or MultiPolygon type");
        }
        else {
            const polygonVector = getGeometryVector(table, EXTENSION_NAME.POLYGON);
            if (polygonVector !== null) {
                return this._renderLayers(polygonVector);
            }
            const multiPolygonVector = getGeometryVector(table, EXTENSION_NAME.MULTIPOLYGON);
            if (multiPolygonVector !== null) {
                return this._renderLayers(multiPolygonVector);
            }
        }
        throw new Error("geometryColumn not Polygon or MultiPolygon");
    }
    // NOTE: Here we shouldn't need a split for handling both multi- and single-
    // geometries, because the underlying SolidPolygonLayer and PathLayer both
    // support multi-* and single- geometries.
    _renderLayers(geometryColumn) {
        const { data: table } = this.props;
        let getPath;
        if (ga.vector.isPolygonVector(geometryColumn)) {
            getPath = getPolygonExterior(geometryColumn);
        }
        else if (ga.vector.isMultiPolygonVector(geometryColumn)) {
            getPath = getMultiPolygonExterior(geometryColumn);
        }
        else {
            assert(false);
        }
        // Layer composition props
        const { data, _dataDiff, stroked, filled, extruded, wireframe, _normalize, _windingOrder, elevationScale, transitions, positionFormat, } = this.props;
        // Rendering props underlying layer
        const { lineWidthUnits, lineWidthScale, lineWidthMinPixels, lineWidthMaxPixels, lineJointRounded, lineMiterLimit, lineDashJustified, } = this.props;
        // Accessor props for underlying layers
        const { getFillColor, getLineColor, getLineWidth, getElevation, getPolygon, updateTriggers, material, } = this.props;
        const FillLayer = this.getSubLayerClass("fill", GeoArrowSolidPolygonLayer);
        const StrokeLayer = this.getSubLayerClass("stroke", GeoArrowPathLayer);
        // Filled Polygon Layer
        const polygonLayer = new FillLayer({
            // _dataDiff,
            extruded,
            elevationScale,
            filled,
            wireframe,
            _normalize,
            _windingOrder,
            getElevation,
            getFillColor,
            getLineColor: extruded && wireframe ? getLineColor : defaultLineColor,
            material,
            transitions,
        }, this.getSubLayerProps({
            id: "fill",
            updateTriggers: updateTriggers && {
                getPolygon: updateTriggers.getPolygon,
                getElevation: updateTriggers.getElevation,
                getFillColor: updateTriggers.getFillColor,
                getLineColor: updateTriggers.getLineColor,
            },
        }), {
            data,
            positionFormat,
            getPolygon,
        });
        // Polygon line layer
        const polygonLineLayer = !extruded &&
            stroked &&
            new StrokeLayer({
                // _dataDiff,
                widthUnits: lineWidthUnits,
                widthScale: lineWidthScale,
                widthMinPixels: lineWidthMinPixels,
                widthMaxPixels: lineWidthMaxPixels,
                jointRounded: lineJointRounded,
                miterLimit: lineMiterLimit,
                dashJustified: lineDashJustified,
                // Already normalized, and since they had been polygons, we know that
                // the lines are a loop.
                _pathType: "loop",
                transitions: transitions && {
                    getWidth: transitions.getLineWidth,
                    getColor: transitions.getLineColor,
                    getPath: transitions.getPolygon,
                },
                getColor: this.getSubLayerAccessor(getLineColor),
                getWidth: this.getSubLayerAccessor(getLineWidth),
            }, this.getSubLayerProps({
                id: "stroke",
                updateTriggers: updateTriggers && {
                    getWidth: updateTriggers.getLineWidth,
                    getColor: updateTriggers.getLineColor,
                    getDashArray: updateTriggers.getLineDashArray,
                },
            }), {
                data: table,
                positionFormat,
                getPath,
                // We only pick solid polygon layers, not the path layers
                pickable: false,
            });
        const layers = [
            // If not extruded: flat fill layer is drawn below outlines
            !extruded && polygonLayer,
            polygonLineLayer,
            // If extruded: draw fill layer last for correct blending behavior
            extruded && polygonLayer,
        ];
        return layers;
    }
}
