// deck.gl-community
// SPDX-License-Identifier: MIT
// Copyright (c) vis.gl contributors
import { CompositeLayer, assert, } from "@deck.gl/core";
import { TextLayer } from "@deck.gl/layers";
import * as ga from "@geoarrow/geoarrow-js";
import { assignAccessor, convertStructToFixedSizeList, expandArrayToCoords, extractAccessorsFromProps, getGeometryVector, isGeomSeparate, } from "../utils/utils";
import { computeChunkOffsets, getPickingInfo, } from "../utils/picking";
import { EXTENSION_NAME } from "../constants";
import { validateAccessors } from "../utils/validate";
// Remove data and getPosition from the upstream default props
const { data: _data, getPosition: _getPosition, getText: _getText, getTextAnchor: _getTextAnchor, getAlignmentBaseline: _getAlignmentBaseline, getPixelOffset: _getPixelOffset, ..._defaultProps } = TextLayer.defaultProps;
// Default props added by us
const ourDefaultProps = {
    getTextAnchor: "middle",
    getAlignmentBaseline: "center",
    getPixelOffset: [0, 0],
    _validate: true,
};
// @ts-expect-error Type error in merging default props with ours
const defaultProps = {
    ..._defaultProps,
    ...ourDefaultProps,
};
export class GeoArrowTextLayer extends CompositeLayer {
    static defaultProps = defaultProps;
    static layerName = "GeoArrowTextLayer";
    getPickingInfo(params) {
        return getPickingInfo(params, this.props.data);
    }
    renderLayers() {
        const { data: table } = this.props;
        if (this.props.getPosition !== undefined) {
            const geometryColumn = this.props.getPosition;
            if (geometryColumn !== undefined &&
                ga.vector.isPointVector(geometryColumn)) {
                return this._renderLayersPoint(geometryColumn);
            }
            throw new Error("getPosition should pass in an arrow Vector of Point type");
        }
        else {
            const pointVector = getGeometryVector(table, EXTENSION_NAME.POINT);
            if (pointVector !== null) {
                return this._renderLayersPoint(pointVector);
            }
        }
        throw new Error("getPosition not GeoArrow point");
    }
    _renderLayersPoint(geometryColumn) {
        const { data: table } = this.props;
        if (this.props._validate) {
            assert(ga.vector.isPointVector(geometryColumn));
            validateAccessors(this.props, table);
        }
        // Exclude manually-set accessors
        const [accessors, otherProps] = extractAccessorsFromProps(this.props, [
            "getPosition",
            "getText",
        ]);
        const tableOffsets = computeChunkOffsets(table.data);
        const layers = [];
        for (let recordBatchIdx = 0; recordBatchIdx < table.batches.length; recordBatchIdx++) {
            let geometryData = geometryColumn.data[recordBatchIdx];
            if (isGeomSeparate(geometryData)) {
                geometryData = convertStructToFixedSizeList(geometryData);
            }
            const flatCoordsData = ga.child.getPointChild(geometryData);
            const flatCoordinateArray = flatCoordsData.values;
            const textData = this.props.getText.data[recordBatchIdx];
            // console.log(textData);
            const textValues = textData.values;
            const characterOffsets = textData.valueOffsets;
            const props = {
                // Note: because this is a composite layer and not doing the rendering
                // itself, we still have to pass in our defaultProps
                ...ourDefaultProps,
                ...otherProps,
                // used for picking purposes
                recordBatchIdx,
                tableOffsets,
                id: `${this.props.id}-geoarrow-heatmap-${recordBatchIdx}`,
                data: {
                    // @ts-expect-error passed through to enable use by function accessors
                    data: table.batches[recordBatchIdx],
                    length: geometryData.length,
                    startIndices: characterOffsets,
                    attributes: {
                        // Positions need to be expanded to be one per character!
                        getPosition: {
                            value: expandArrayToCoords(flatCoordinateArray, geometryData.type.listSize, characterOffsets),
                            size: geometryData.type.listSize,
                        },
                        // TODO: support non-ascii characters
                        getText: {
                            value: textValues,
                            // size: 1,
                        },
                    },
                },
            };
            for (const [propName, propInput] of Object.entries(accessors)) {
                assignAccessor({
                    props,
                    propName,
                    propInput,
                    chunkIdx: recordBatchIdx,
                    geomCoordOffsets: characterOffsets,
                });
            }
            const layer = new TextLayer(this.getSubLayerProps(props));
            layers.push(layer);
        }
        return layers;
    }
}
